"""The base [`Optimizer`][amltk.optimization.optimizer.Optimizer] class,
defines the API we require optimizers to implement.

* [`ask()`][amltk.optimization.optimizer.Optimizer.ask] - Ask the optimizer for a
    new [`Trial`][amltk.optimization.trial.Trial] to evaluate.
* [`tell()`][amltk.optimization.optimizer.Optimizer.tell] - Tell the optimizer
    the result of the sampled config. This comes in the form of a
    [`Trial.Report`][amltk.optimization.trial.Trial.Report].

Additionally, to aid users from switching between optimizers, the
[`preferred_parser()`][amltk.optimization.optimizer.Optimizer.preferred_parser]
method should return either a `parser` function or a string that can be used
with [`node.search_space(parser=..._)`][amltk.pipeline.Node.search_space] to
extract the search space for the optimizer.
"""
from __future__ import annotations

import logging
from abc import abstractmethod
from collections.abc import Callable, Iterable, Iterator, Sequence
from datetime import datetime
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Concatenate,
    Generic,
    ParamSpec,
    Protocol,
    TypeVar,
    overload,
)
from typing_extensions import Self

from more_itertools import all_unique

from amltk.store.paths.path_bucket import PathBucket

if TYPE_CHECKING:
    from amltk.optimization.metric import Metric
    from amltk.optimization.trial import Trial
    from amltk.pipeline import Node
    from amltk.types import Seed

I = TypeVar("I")  # noqa: E741
P = ParamSpec("P")
ParserOutput = TypeVar("ParserOutput")

logger = logging.getLogger(__name__)


class Optimizer(Generic[I]):
    """An optimizer protocol.

    An optimizer is an object that can be asked for a trail using `ask` and a
    `tell` to inform the optimizer of the report from that trial.
    """

    metrics: Sequence[Metric]
    """The metrics to optimize."""

    bucket: PathBucket
    """The bucket to give to trials generated by this optimizer."""

    def __init__(
        self,
        metrics: Sequence[Metric],
        bucket: PathBucket | None = None,
    ) -> None:
        """Initialize the optimizer.

        Args:
            metrics: The metrics to optimize.
            bucket: The bucket to store results of individual trials from this
                optimizer.
        """
        super().__init__()
        if not all_unique(metric.name for metric in metrics):
            raise ValueError(
                "All metrics must have unique names."
                f"Got {metrics} with names {[metric.name for metric in metrics]}",
            )

        self.metrics = metrics
        self.bucket = (
            bucket
            if bucket is not None
            else PathBucket(f"{self.__class__.__name__}-{datetime.now().isoformat()}")
        )

    @abstractmethod
    def tell(self, report: Trial.Report[I]) -> None:
        """Tell the optimizer the report for an asked trial.

        Args:
            report: The report for a trial
        """

    @overload
    @abstractmethod
    def ask(self, n: int) -> Iterable[Trial[I]]:
        ...

    @overload
    @abstractmethod
    def ask(self, n: None = None) -> Trial[I]:
        ...

    @abstractmethod
    def ask(self, n: int | None = None) -> Trial[I] | Iterable[Trial[I]]:
        """Ask the optimizer for a trial to evaluate.

        Args:
            n: The number of trials to ask for. If `None`, ask for a single trial.

        Returns:
            A config to sample.
        """
        ...

    @classmethod
    def preferred_parser(
        cls,
    ) -> str | Callable[Concatenate[Node, ...], Any] | Callable[[Node], Any] | None:
        """The preferred parser for this optimizer.

        !!! note

            Subclasses should override this as required.

        """
        return None

    @classmethod
    @abstractmethod
    def create(
        cls,
        *,
        space: Node,
        metrics: Metric | Sequence[Metric],
        bucket: str | Path | PathBucket | None = None,
        seed: Seed | None = None,
    ) -> Self:
        """Create this optimizer.

        !!! note

            Subclasses should override this with more specific configuration
            but these arguments should be all that's necessary to create the optimizer.

        Args:
            space: The space to optimize over.
            bucket: The bucket for where to store things related to the trial.
            metrics: The metrics to optimize.
            seed: The seed to use for the optimizer.

        Returns:
            The optimizer.
        """

    class CreateSignature(Protocol):
        """A Protocol which defines the keywords required to create an
        optimizer with deterministic behavior at a desired location.

        This protocol matches the `Optimizer.create` classmethod, however we also
        allow any function which accepts the keyword arguments to create an
        Optimizer.
        """

        def __call__(
            self,
            *,
            space: Node,
            metrics: Metric | Sequence[Metric],
            bucket: PathBucket | None = None,
            seed: Seed | None = None,
        ) -> Optimizer:
            """A function which creates an optimizer for node.optimize should
            accept the following keyword arguments.

            Args:
                space: The node to optimize
                metrics: The metrics to optimize
                bucket: The bucket to store the results in
                seed: The seed to use for the optimization
            """
            ...

    @classmethod
    def _get_known_importable_optimizer_classes(cls) -> Iterator[type[Optimizer]]:
        """Get all developer known optimizer classes. This is used for defaults.

        Do not rely on this functionality and prefer to give concrete optimizers to
        functionality requiring one. This is intended for convenience of particular
        quickstart methods.
        """
        # NOTE: We can't use the `Optimizer.__subclasses__` method as the optimizers
        # are not imported by any other module initially and so they do no exist
        # until imported. Hence this manual iteration. For now, we be explicit and
        # only if the optimizer list grows should we consider dynamic importing.
        try:
            from amltk.optimization.optimizers.smac import SMACOptimizer

            yield SMACOptimizer
        except ImportError as e:
            logger.debug("Failed to import SMACOptimizer", exc_info=e)

        try:
            from amltk.optimization.optimizers.optuna import OptunaOptimizer

            yield OptunaOptimizer
        except ImportError as e:
            logger.debug("Failed to import OptunaOptimizer", exc_info=e)

        try:
            from amltk.optimization.optimizers.neps import NEPSOptimizer

            yield NEPSOptimizer
        except ImportError as e:
            logger.debug("Failed to import NEPSOptimizer", exc_info=e)



from amltk._richutil.renderers._make_column_selector import rich_make_column_selector
from amltk._richutil.renderers.executors import ProcessPoolExecutorRenderer
from amltk._richutil.renderers.function import Function

__all__ = ["Function", "rich_make_column_selector", "ProcessPoolExecutorRenderer"]



"""[Optuna](https://optuna.org/) parser for parsing out a
[`search_space()`][amltk.pipeline.node.Node.search_space].
from a pipeline.

!!! tip "Requirements"

    This requires `Optuna` which can be installed with:

    ```bash
    pip install amltk[optuna]

    # Or directly
    pip install optuna
    ```

??? warning "Limitations"

    Optuna feature a very dynamic search space (_define-by-run_),
    where people typically sample from some trial object and use traditional
    python control flow to define conditionality.

    This means we can not trivially represent this conditionality in a static
    search space. While _band-aids_ are possible,
    it naturally does not sit well with the static output of a parser.

    As such, our parser **does not support conditionals or choices!**.
    Users may still use the _define-by-run_ within their optimization function
    itself.

    If you have experience with Optuna and have any suggestions,
    please feel free to open an issue or PR on GitHub!

### Usage
The typical way to represent a search space for Optuna is just to use a dictionary,
where the keys are the names of the hyperparameters and the values are either
integer/float tuples indicating boundaries or some discrete set of values.
It is possible to have the value directly be a
`BaseDistribution`, an optuna type, when you need to customize the distribution more.


```python exec="true" source="material-block" html="true" session="optuna-parser"
from amltk.pipeline import Component
from optuna.distributions import FloatDistribution

c = Component(
    object,
    space={
        "myint": (1, 10),
        "myfloat": (1.0, 10.0),
        "mycategorical": ["a", "b", "c"],
        "log-scale-custom": FloatDistribution(1e-10, 1e-2, log=True),
    },
    name="name",
)

space = c.search_space(parser="optuna")
from amltk._doc import doc_print; doc_print(print, space)  # markdown-exec: hide
```

You may also just pass the `parser=` function directly if preferred

```python exec="true" source="material-block" html="true" session="optuna-parser"
from amltk.pipeline.parsers.optuna import parser as optuna_parser

space = c.search_space(parser=optuna_parser)
from amltk._doc import doc_print; doc_print(print, space)  # markdown-exec: hide
```

When using [`search_space()`][amltk.pipeline.node.Node.search_space] on a some nested
structures, you may want to flatten the names of the hyperparameters. For this you
can use `flat=`

```python exec="true" source="material-block" html="true"
from amltk.pipeline import Searchable, Sequential

seq = Sequential(
    Searchable({"myint": (1, 10)}, name="nested_1"),
    Searchable({"myfloat": (1.0, 10.0)}, name="nested_2"),
    name="seq"
)

hierarchical_space = seq.search_space(parser="optuna", flat=False)  # Default
from amltk._doc import doc_print; doc_print(print, hierarchical_space)  # markdown-exec: hide

flat_space = seq.search_space(parser="optuna", flat=False)  # Default
from amltk._doc import doc_print; doc_print(print, flat_space)  # markdown-exec: hide
```

"""  # noqa: E501

from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING

import numpy as np
from optuna.distributions import (
    BaseDistribution,
    CategoricalChoiceType,
    CategoricalDistribution,
    FloatDistribution,
    IntDistribution,
)

from amltk._functional import prefix_keys

if TYPE_CHECKING:
    from typing import TypeAlias

    from amltk.pipeline import Node

    OptunaSearchSpace: TypeAlias = dict[str, BaseDistribution]

PAIR = 2


def _convert_hp_to_optuna_distribution(
    name: str,
    hp: tuple | Sequence | CategoricalChoiceType | BaseDistribution,
) -> BaseDistribution:
    match hp:
        case BaseDistribution():
            return hp
        case None | bool() | int() | str() | float():
            return CategoricalDistribution([hp])
        case tuple() as tup if len(tup) == PAIR:
            match tup:
                case (int() | np.integer(), int() | np.integer()):
                    x, y = tup
                    return IntDistribution(int(x), int(y))
                case (float() | np.floating(), float() | np.floating()):
                    x, y = tup
                    return FloatDistribution(float(x), float(y))
                case (x, y):
                    raise ValueError(
                        f"Expected {name} to have same type for lower and upper bound,"
                        f"got lower: {type(x)}, upper: {type(y)}.",
                    )
        case Sequence():
            if len(hp) == 0:
                raise ValueError(f"Can't have empty list for categorical {name}")

            return CategoricalDistribution(hp)
        case _:
            raise ValueError(
                f"Could not parse {name} as a valid Optuna distribution." f"\n{hp=}",
            )

    raise ValueError(f"Could not parse {name} as a valid Optuna distribution.\n{hp=}")


def _parse_space(node: Node) -> OptunaSearchSpace:
    match node.space:
        case None:
            space = {}
        case Mapping():
            space = {
                name: _convert_hp_to_optuna_distribution(name=name, hp=hp)
                for name, hp in node.space.items()
            }
        case _:
            raise ValueError(
                f"Can only parse mappings with Optuna but got {node.space=}",
            )

    if node.config is not None:
        for name, value in node.config.items():
            if name in space:
                space[name] = CategoricalDistribution([value])

    return space


def parser(
    node: Node,
    *,
    flat: bool = False,
    conditionals: bool = False,
    delim: str = ":",
) -> OptunaSearchSpace:
    """Parse a Node and its children into a ConfigurationSpace.

    Args:
        node: The Node to parse
        flat: Whether to have a hierarchical naming scheme for nodes and their children.
        conditionals: Whether to include conditionals in the space from a
            [`Choice`][amltk.pipeline.Choice]. If this is `False`, this will
            also remove all forbidden clauses and other conditional clauses.
            The primary use of this functionality is that some optimizers do not
            support these features.

            !!! TODO "Not yet supported"

                This functionality is not yet supported as we can't encode this into
                a static Optuna search space.

        delim: The delimiter to use for the names of the hyperparameters.
    """
    if conditionals:
        raise NotImplementedError("Conditionals are not yet supported with Optuna.")

    space = prefix_keys(_parse_space(node), prefix=f"{node.name}{delim}")

    for child in node.nodes:
        subspace = parser(child, flat=flat, conditionals=conditionals, delim=delim)
        if not flat:
            subspace = prefix_keys(subspace, prefix=f"{node.name}{delim}")

        for name, hp in subspace.items():
            if name in space:
                raise ValueError(
                    f"Duplicate name {name} already in space from space of {node.name}",
                    f"\nCurrently parsed space: {space}",
                )
            space[name] = hp

    return space



"""Stores low-level types used through the library."""
from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from itertools import chain, repeat
from typing import Any, NoReturn, Protocol, TypeAlias, TypeVar
from typing_extensions import override

import numpy as np

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

# NOTE: We do not provide sorted types for mutable data structures
# as modifying them could potentially ruin the sorting.

SortedSequence: TypeAlias = Sequence[T]
"""A sequence that is sorted. Only useful for typing"""

SortedIterable: TypeAlias = Iterable[T]
"""An iterable that is sorted. Only useful for typing"""

Item = TypeVar("Item")
"""The type associated with components, splits and choices"""

Config: TypeAlias = Mapping[str, Any]
"""An object representing a configuration of a pipeline."""

Space = TypeVar("Space")
"""Generic for objects that are aware of a space but not the specific kind"""

Seed: TypeAlias = int | np.integer | np.random.RandomState | np.random.Generator
"""Type alias for kinds of Seeded objects."""

FidT: TypeAlias = tuple[int, int] | tuple[float, float] | list[Any]
"""Type alias for a fidelity bound."""


class Comparable(Protocol):
    """Protocol for annotating comparable types."""

    @abstractmethod
    def __lt__(self: _CT, other: _CT, /) -> bool:
        pass

    @abstractmethod
    def __gt__(self: _CT, other: _CT, /) -> bool:
        pass


_CT = TypeVar("_CT", bound=Comparable)


def assert_never(value: NoReturn) -> NoReturn:
    """Utility function for asserting that a value is never reached."""
    # This also works in runtime as well:
    raise AssertionError(f"This code should never be reached, got: {value}")


def safe_issubclass(cls: type, classes: str | tuple[str, ...]) -> bool:
    """Check if a class is a subclass of a given type.

    This is a safe version of issubclass that relies on strings,
    which is useful for when the type is not importable.

    Args:
        cls: The class to check
        classes: The type to check for.

    Returns:
        bool
    """

    def type_names(o: type) -> Iterator[str]:
        yield o.__qualname__
        for parent in o.__bases__:
            yield from type_names(parent)

    allowable_names = {classes} if isinstance(classes, str) else set(classes)
    return any(name in allowable_names for name in type_names(cls))


def safe_isinstance(obj: Any, t: str | tuple[str, ...]) -> bool:
    """Check if an object is of a given type.

    This is a safe version of isinstance that relies on strings,
    which is useful for when the type is not importable.

    Args:
        obj: The object to check
        t: The type to check for.

    Returns:
        bool
    """
    return safe_issubclass(type(obj), t)


class Requeue(Iterator[T]):
    """A queue that can have items requeued.

    ```python exec="true" source="material-block" result="python" title="Requeue"
    import random
    from amltk.types import Requeue

    name_generator = iter(["Alice", "Bob", "Charlie"])
    queue: Requeue[str] = Requeue(name_generator)

    rng = random.Random(1)

    def process_name(name: str) -> bool:
        return rng.choice([True, False])

    for name in queue:
        print(f"Processing {name}")
        processed = process_name(name)
        if not processed:
            print(f"Failed to process {name}, requeuing")
            queue.requeue(name)
    ```


    See Also:
        * [`Requeue.from_func(f)`][amltk.types.Requeue.from_func]

            If you have a function which will generate items, you can use
            this to create a requeue from it.

        * [`.append(item)`][amltk.types.Requeue.append]

            Append an item to the end of the queue

        * [`.requeue(item)`][amltk.types.Requeue.requeue]

            Requeue an item to the start of the queue
    """

    def __init__(self, generator: Iterable[T]) -> None:
        """Create a requeue from an iterable."""
        super().__init__()
        self.generator = iter(generator)

    @override
    def __next__(self) -> T:
        return next(self.generator)

    def append(self, item: T) -> None:
        """Append an item to the queue."""
        self.generator = chain(self.generator, [item])

    def requeue(self, item: T) -> None:
        """Requeue an item."""
        self.generator = chain([item], self.generator)

    @classmethod
    def from_func(cls, func: Callable[[], T], n: int | None = None) -> Requeue[T]:
        """Create a Requeue from a function."""
        repeater = repeat(None) if n is None else repeat(None, times=n)
        return cls(func() for _ in repeater)



"""The [`PynisherPlugin`][amltk.scheduling.plugins.pynisher.PynisherPlugin]
uses [pynisher](https://github.com/automl/pynisher) to place **memory**, **walltime**
and **cputime** constraints on processes, crashing them if these limits are reached.
These default units are `bytes ("B")` and `seconds ("s")` but you can also use other
units, please see the relevant API doc.

It's best use is when used with
[`Scheduler.with_processes()`][amltk.scheduling.Scheduler.with_processes] to have work
performed in processes.

!!! tip "Requirements"

    This required `pynisher` which can be installed with:

    ```bash
    pip install amltk[pynisher]

    # Or directly
    pip install pynisher
    ```

??? tip "Usage"

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Task, Scheduler
    from amltk.scheduling.plugins.pynisher import PynisherPlugin
    import time

    def f(x: int) -> int:
        time.sleep(x)
        return 42

    scheduler = Scheduler.with_processes()
    task = scheduler.task(f, plugins=PynisherPlugin(walltime_limit=(1, "s")))

    @task.on("pynisher-timeout")
    def callback(exception):
        pass
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```

??? example "`@events`"

    === "`@pynisher-timeout`"

        ::: amltk.scheduling.plugins.pynisher.PynisherPlugin.TIMEOUT

    === "`@pynisher-memory-limit`"

        ::: amltk.scheduling.plugins.pynisher.PynisherPlugin.MEMORY_LIMIT_REACHED

    === "`@pynisher-cputime-limit`"

        ::: amltk.scheduling.plugins.pynisher.PynisherPlugin.CPU_TIME_LIMIT_REACHED

    === "`@pynisher-walltime-limit`"

        ::: amltk.scheduling.plugins.pynisher.PynisherPlugin.WALL_TIME_LIMIT_REACHED

??? warning "Scheduler Executor"

    This will place process limits on the task as soon as it starts
    running, whever it may be running. If you are using
    [`Scheduler.with_sequential()`][amltk.Scheduler.with_sequential]
    then this will place limits on the main process, likely not what you
    want. This also does not work with a
    [`ThreadPoolExecutor`][concurrent.futures.ThreadPoolExecutor].

    If using this with something like [`dask-jobqueue`],
    then this will place limits on the workers it spawns. It would be better
    to place limits directly through dask job-queue then.

??? warning "Platform Limitations (Mac, Windows)"

    Pynisher has some limitations with memory on Mac and Windows:
    https://github.com/automl/pynisher#features

    You can check this with `PynisherPlugin.supports("memory")`,
    `PynisherPlugin.supports("cpu_time")` and
    `PynisherPlugin.supports("wall_time")`.
    See [`PynisherPlugin.supports()`][amltk.scheduling.plugins.pynisher.PynisherPlugin.supports]
"""  # noqa: E501
from __future__ import annotations

import traceback
from collections.abc import Callable
from dataclasses import dataclass
from multiprocessing.context import BaseContext
from typing import TYPE_CHECKING, ClassVar, Generic, Literal, TypeAlias, TypeVar
from typing_extensions import ParamSpec, Self, override

import pynisher
import pynisher.exceptions

from amltk.optimization import Trial
from amltk.scheduling.events import Event
from amltk.scheduling.plugins.plugin import Plugin

if TYPE_CHECKING:
    import asyncio

    from rich.panel import Panel

    from amltk.scheduling.task import Task

P = ParamSpec("P")
R = TypeVar("R")


@dataclass
class _PynisherWrap(Generic[P, R]):
    fn: Callable[P, R]
    memory_limit: int | tuple[int, str] | None = None
    cputime_limit: int | tuple[float, str] | None = None
    walltime_limit: int | tuple[float, str] | None = None
    terminate_child_processes: bool = True
    context: BaseContext | None = None
    disable_trial_handling: bool = False

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        if any(
            limit is not None
            for limit in (self.memory_limit, self.cputime_limit, self.walltime_limit)
        ):
            fn = pynisher.Pynisher(
                self.fn,
                memory=self.memory_limit,
                cpu_time=self.cputime_limit,
                wall_time=self.walltime_limit,
                terminate_child_processes=True,
                context=self.context,
            )
        else:
            fn = self.fn

        trial: Trial | None = None
        if not self.disable_trial_handling:
            if len(args) > 0 and isinstance(args[0], Trial):
                trial = args[0]
            elif (_trial := kwargs.get("trial")) is not None:
                if not isinstance(_trial, Trial):
                    raise ValueError(
                        f"Expected 'trial' to be a Trial instance, got {type(trial)}"
                        f"\n{trial=}",
                    )
                trial = _trial

        if trial is not None:
            try:
                return fn(*args, **kwargs)
            except pynisher.PynisherException as e:
                tb = traceback.format_exc()
                return trial.fail(e, tb)  # type: ignore
        else:
            return fn(*args, **kwargs)


class PynisherPlugin(Plugin):
    """A plugin that wraps a task in a pynisher to enforce limits on it.

    This plugin wraps a task function in a `Pynisher` instance to enforce
    limits on the task. The limits are set by any of `memory_limit=`,
    `cpu_time_limit=` and `wall_time_limit=`.

    Adds four new events to the task

    * [`@pynisher-timeout`][amltk.scheduling.plugins.pynisher.PynisherPlugin.TIMEOUT]
    * [`@pynisher-memory-limit`][amltk.scheduling.plugins.pynisher.PynisherPlugin.MEMORY_LIMIT_REACHED]
    * [`@pynisher-cputime-limit`][amltk.scheduling.plugins.pynisher.PynisherPlugin.CPU_TIME_LIMIT_REACHED]
    * [`@pynisher-walltime-limit`][amltk.scheduling.plugins.pynisher.PynisherPlugin.WALL_TIME_LIMIT_REACHED]

    Attributes:
        memory_limit: The memory limit of the task.
        cpu_time_limit: The cpu time limit of the task.
        wall_time_limit: The wall time limit of the task.
    """  # noqa: E501

    name: ClassVar = "pynisher-plugin"
    """The name of the plugin."""

    TIMEOUT: Event[PynisherPlugin.TimeoutException] = Event("pynisher-timeout")
    """A Task timed out, either due to the wall time or cpu time limit.

    Will call any subscribers with the exception as the argument.

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Task, Scheduler
    from amltk.scheduling.plugins.pynisher import PynisherPlugin
    import time

    def f(x: int) -> int:
        time.sleep(x)
        return 42

    scheduler = Scheduler.with_processes()
    task = scheduler.task(f, plugins=PynisherPlugin(walltime_limit=(1, "s")))

    @task.on("pynisher-timeout")
    def callback(exception):
        pass
    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
    """

    MEMORY_LIMIT_REACHED: Event[pynisher.exceptions.MemoryLimitException] = Event(
        "pynisher-memory-limit",
    )
    """A Task was submitted but reached it's memory limit.

    Will call any subscribers with the exception as the argument.

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Task, Scheduler
    from amltk.scheduling.plugins.pynisher import PynisherPlugin
    import numpy as np

    def f(x: int) -> int:
        x = np.arange(100000000)
        time.sleep(x)
        return 42

    scheduler = Scheduler.with_processes()
    task = scheduler.task(f, plugins=PynisherPlugin(memory_limit=(1, "KB")))

    @task.on("pynisher-memory-limit")
    def callback(exception):
        pass

    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
    """

    CPU_TIME_LIMIT_REACHED: Event[pynisher.exceptions.CpuTimeoutException] = Event(
        "pynisher-cputime-limit",
    )
    """A Task was submitted but reached it's cpu time limit.

    Will call any subscribers with the exception as the argument.

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Task, Scheduler
    from amltk.scheduling.plugins.pynisher import PynisherPlugin
    import time

    def f(x: int) -> int:
        i = 0
        while True:
            # Keep busying computing the answer to everything
            i += 1

        return 42

    scheduler = Scheduler.with_processes()
    task = scheduler.task(f, plugins=PynisherPlugin(cputime_limit=(1, "s")))

    @task.on("pynisher-cputime-limit")
    def callback(exception):
        pass

    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
    """

    WALL_TIME_LIMIT_REACHED: Event[pynisher.exceptions.WallTimeoutException] = Event(
        "pynisher-walltime-limit",
    )
    """A Task was submitted but reached it's wall time limit.

    Will call any subscribers with the exception as the argument.

    ```python exec="true" source="material-block" html="true"
    from amltk.scheduling import Task, Scheduler
    from amltk.scheduling.plugins.pynisher import PynisherPlugin
    import time

    def f(x: int) -> int:
        time.sleep(x)
        return 42

    scheduler = Scheduler.with_processes()
    task = scheduler.task(f, plugins=PynisherPlugin(walltime_limit=(1, "s")))

    @task.on("pynisher-walltime-limit")
    def callback(exception):
        pass

    from amltk._doc import doc_print; doc_print(print, task)  # markdown-exec: hide
    ```
    """

    PynisherException: TypeAlias = pynisher.exceptions.PynisherException
    """The base exception for all pynisher exceptions."""

    TimeoutException: TypeAlias = pynisher.exceptions.TimeoutException
    """The exception that is raised when a task times out."""

    MemoryLimitException: TypeAlias = pynisher.exceptions.MemoryLimitException
    """The exception that is raised when a task reaches it's memory limit."""

    CpuTimeoutException: TypeAlias = pynisher.exceptions.CpuTimeoutException
    """The exception that is raised when a task reaches it's cpu time limit."""

    WallTimeoutException: TypeAlias = pynisher.exceptions.WallTimeoutException
    """The exception that is raised when a task reaches it's wall time limit."""

    def __init__(
        self,
        *,
        memory_limit: int | tuple[int, str] | None = None,
        cputime_limit: int | tuple[float, str] | None = None,
        walltime_limit: int | tuple[float, str] | None = None,
        context: BaseContext | None = None,
        disable_trial_handling: bool = False,
    ):
        """Initialize a `PynisherPlugin` instance.

        Args:
            memory_limit: The memory limit to wrap the task in. Base unit is in bytes
                but you can specify `(value, unit)` where `unit` is one of
                `("B", "KB", "MB", "GB")`. Defaults to `None`
            cputime_limit: The cpu time limit to wrap the task in. Base unit is in
                seconds but you can specify `(value, unit)` where `unit` is one of
                `("s", "m", "h")`. Defaults to `None`
            walltime_limit: The wall time limit for the task. Base unit is in seconds
                but you can specify `(value, unit)` where `unit` is one of
                `("s", "m", "h")`. Defaults to `None`.
            context: The context to use for multiprocessing. Defaults to `None`.
                See [`multiprocessing.get_context()`][multiprocessing.get_context]
            disable_trial_handling: By default, the `PynisherPlugin` will auto-detect
                if the task is one for a `Trial`. If so, it will catch any pynisher
                specific exceptions and return a `Trial.Report` with `trial.fail()`,
                instead of raising the expcetion. This has the effect that the report
                can be caught with `task.on_result` where it can be handled. This will
                also prevent the specific events `@pynisher-timeout`,
                `@pynisher-memory-limit`, `@pynisher-cputime-limit`
                and `@pynisher-walltime-limit` from being emitted.

                If this
                is `True`, then the pynisher exceptions will be raised as normal and
                should be handled with `task.on_exception` where there is no direct
                access to the `Trial` submitted.

                ??? note "Auto-Detection"

                    This will be triggered if the first positional argument is a
                    `Trial` or if any of the keyword arguments are `"trial"`.

                    ```python
                    from amltk.optimization import Trial
                    from amltk.scheduling import Scheduler

                    def trial_evaluator_one(trial: Trial, ...) -> int:
                        ...

                    def trial_evaluator_two(..., trial: Trial) -> int:
                        ...

                    scheduler = Scheduler.with_processes(1)

                    task_one = scheduler.task(
                        trial_evaluator_one,
                        plugins=PynisherPlugin(memory_limit=(1, "gb")
                    )
                    task_two = scheduler.task(
                        trial_evaluator_two,
                        plugins=PynisherPlugin(memory_limit=(1, "gb")
                    )

                    # Will auto-detect
                    trial = Trial(...)
                    task_one.submit(trial, ...)
                    task_two.submit(..., trial=trial)

                    # Will not auto-detect
                    task_one.submit(42, trial)
                    ```
        """
        super().__init__()

        for limit, name in [
            (memory_limit, "memory"),
            (cputime_limit, "cpu_time"),
            (walltime_limit, "wall_time"),
        ]:
            if limit is not None and not self.supports(name):  # type: ignore
                raise RuntimeError(
                    f"Your platform does not support {name} limits."
                    " Please see pynisher documentation for more:"
                    "\nhttps://github.com/automl/pynisher#features",
                )

        self.memory_limit = memory_limit
        self.cputime_limit = cputime_limit
        self.walltime_limit = walltime_limit
        self.context = context
        self.disable_trial_handling = disable_trial_handling

        self.task: Task

    @override
    def pre_submit(
        self,
        fn: Callable[P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> tuple[Callable[P, R], tuple, dict]:
        """Wrap a task function in a `Pynisher` instance."""
        _fn = _PynisherWrap(
            fn,
            disable_trial_handling=self.disable_trial_handling,
            memory_limit=self.memory_limit,
            cputime_limit=self.cputime_limit,
            walltime_limit=self.walltime_limit,
            context=self.context,
            terminate_child_processes=True,
        )
        return _fn, args, kwargs

    @override
    def attach_task(self, task: Task) -> None:
        """Attach the plugin to a task."""
        self.task = task
        task.emitter.add_event(
            self.TIMEOUT,
            self.MEMORY_LIMIT_REACHED,
            self.CPU_TIME_LIMIT_REACHED,
            self.WALL_TIME_LIMIT_REACHED,
        )

        # Check the exception and emit pynisher specific ones too
        task.on_exception(self._check_to_emit_pynisher_exception, hidden=True)

    @override
    def copy(self) -> Self:
        """Return a copy of the plugin.

        Please see [`Plugin.copy()`][amltk.Plugin.copy].
        """
        return self.__class__(
            memory_limit=self.memory_limit,
            cputime_limit=self.cputime_limit,
            walltime_limit=self.walltime_limit,
        )

    def _check_to_emit_pynisher_exception(
        self,
        _: asyncio.Future,
        exception: BaseException,
    ) -> None:
        """Check if the exception is a pynisher exception and emit it."""
        if isinstance(exception, pynisher.CpuTimeoutException):
            self.task.emitter.emit(self.TIMEOUT, exception)
            self.task.emitter.emit(self.CPU_TIME_LIMIT_REACHED, exception)
        elif isinstance(exception, pynisher.WallTimeoutException):
            self.task.emitter.emit(self.TIMEOUT)
            self.task.emitter.emit(self.WALL_TIME_LIMIT_REACHED, exception)
        elif isinstance(exception, pynisher.MemoryLimitException):
            self.task.emitter.emit(self.MEMORY_LIMIT_REACHED, exception)

    @classmethod
    def supports(cls, kind: Literal["wall_time", "cpu_time", "memory"]) -> bool:
        """Check if the task is supported by the plugin.

        Args:
            kind: The kind of limit to check.

        Returns:
            `True` if the limit is supported by the plugin for your os, else `False`.
        """
        return pynisher.supports(kind)

    @override
    def __rich__(self) -> Panel:
        from rich.panel import Panel
        from rich.pretty import Pretty
        from rich.table import Table

        table = Table(
            "Memory",
            "Wall Time",
            "CPU Time",
            padding=(0, 1),
            show_edge=False,
            box=None,
        )
        table.add_row(
            Pretty(self.memory_limit),
            Pretty(self.walltime_limit),
            Pretty(self.cputime_limit),
        )
        return Panel(table, title=f"Plugin {self.name}")



"""Module containing the base definition of a bucket.

A bucket is a collection of resources that can be accessed by a key
of a given type. This lets you easily store and retrieve objects of varying
types in a single location.

???+ note "Concrete examples"

    * [`PathBucket`][amltk.store.paths.path_bucket.PathBucket].
"""
from __future__ import annotations

import re
from abc import ABC, abstractmethod
from collections.abc import (
    Callable,
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
)
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeVar, overload
from typing_extensions import override

from more_itertools import ilen

from amltk.store.drop import Drop

T = TypeVar("T")

if TYPE_CHECKING:
    from typing_extensions import Self

    from amltk.store.loader import Loader

DEFAULT_FILE_LOADERS: tuple[Loader, ...] = ()

LinkT = TypeVar("LinkT")
KeyT = TypeVar("KeyT", bound=Hashable)


class Bucket(ABC, MutableMapping[KeyT, Drop[LinkT]], Generic[KeyT, LinkT]):
    """Definition of a bucket of resources, accessed by a Key.

    Indexing into a bucket returns a [`Drop`][amltk.store.drop.Drop] that
    can be used to access the resource.

    The definition mostly follow that of MutableMapping, but with
    the change of `.keys()` and `.values()` to return iterators
    and `.items()` to return an iterator of tuples.
    The other change is that the `.values()` do not return the
    resources themselves, by rather a [`Drop`][amltk.store.drop.Drop]
    which wraps the resource.
    """

    @override
    @abstractmethod
    def __setitem__(self, key: KeyT, value: Any) -> None:
        """Store a value in the bucket.

        Args:
            key: The key to the resource.
            value: The value to store in the bucket.
        """

    @override
    @abstractmethod
    def __getitem__(self, key: KeyT) -> Drop[LinkT]:
        """Get a drop for a resource in the bucket.

        Args:
            key: The key to the resource.
        """

    @override
    @abstractmethod
    def __delitem__(self, key: KeyT) -> None:
        """Remove a resource from the bucket.

        Args:
            key: The key to the resource.
        """

    @override
    @abstractmethod
    def __iter__(self) -> Iterator[KeyT]:
        """Iterate over the keys in the bucket."""

    @abstractmethod
    def sub(self, key: KeyT) -> Self:
        """Create a subbucket of this bucket.

        Args:
            key: The name of the sub bucket.

        Returns:
            A new bucket with the same loaders as the current bucket.
        """

    @override
    def __contains__(self, key: object) -> bool:
        """Check if a key is in the bucket.

        Args:
            key: The key to check for.
        """
        return key in self

    @override
    def __len__(self) -> int:
        """Get the number of keys in the bucket."""
        return ilen(iter(self))

    @overload
    def find(self, pattern: str) -> dict[str, Drop[LinkT]] | None:
        ...

    @overload
    def find(
        self,
        pattern: str,
        *,
        multi_key: Literal[False] = False,
    ) -> dict[str, Drop[LinkT]] | None:
        ...

    @overload
    def find(
        self,
        pattern: str,
        *,
        multi_key: Literal[True],
    ) -> dict[tuple[str, ...], Drop[LinkT]] | None:
        ...

    @overload
    def find(
        self,
        pattern: str,
        *,
        multi_key: bool,
    ) -> dict[str, Drop[LinkT]] | dict[tuple[str, ...], Drop[LinkT]] | None:
        ...

    def find(
        self,
        pattern: str,
        *,
        multi_key: bool = False,
    ) -> dict[str, Drop[LinkT]] | dict[tuple[str, ...], Drop[LinkT]] | None:
        """Find resources in the bucket.

        ```python
        found = bucket.find(r"trial_(.+)_val_predictions.npy")  # (1)!
        if found is None:
            raise KeyError("No predictions found")

        for name, drop in found.items():
            predictions = drop.get()
            # Do something with the predictions
            # ...
        ```

        1. The `(.+)` is a **capture group** which will attempt to match anything `.`,
            when there is one or more occurences `+`, and put it in a capure group `()`.
            What is captured will be used as the key in the returned dict.

        Args:
            pattern: The pattern to search for.
            multi_key: Whether you have multiple capture groups in the pattern.

                !!! note "Multiple capture groups with `()`"

                    If using multiple capture groups, the returned dict will have
                    tuples as keys. If there is only one capture group, the tuple
                    will be expanded to a single value.

        Returns:
            A mapping of links to drops for the resources found.
        """
        keys = [(key, match) for key in self if (match := re.search(pattern, str(key)))]
        if not keys:
            return None

        matches = {match.groups(): self[key] for key, match in keys}

        # If it's a tuple of length 1, we expand it
        one_group = len(next(iter(matches.keys()))) == 1
        if one_group:
            if multi_key:
                raise ValueError(
                    "Use multi_key=True when the pattern has more than 1 capture group",
                )

            return {key[0]: drop for key, drop in matches.items()}

        # Here we have multi-groups => tuple keys
        if not multi_key:
            raise ValueError(
                "Use multi_key=False when the pattern has only 1 capture group",
            )

        return matches

    def store(self, other: Mapping[KeyT, Any]) -> None:
        """Store items into the bucket with the given mapping.

        Args:
            other: The mapping of items to store in the bucket.
        """
        for key, value in other.items():
            self[key] = value

    def fetch(
        self,
        *keys: KeyT,
        default: None | Any | dict[KeyT, Any] = None,
    ) -> dict[KeyT, Any]:
        """Fetch a resource from the bucket.

        Args:
            keys: The keys to the resources.
            default: The default value to return if the key is not in the bucket.
                If a dict is passed, the default for each key will be the value
                in the dict for that key, using None if not present.

        Returns:
            The resources stored in the bucket at the given keys.
        """
        default_dict = {} if not isinstance(default, dict) else default
        return {
            key: self[key].get(default=default_dict.get(key, default)) for key in keys
        }

    @override
    def update(self, items: Mapping[KeyT, Any]) -> None:  # type: ignore
        """Update the bucket with the given mapping.

        Args:
            items: The mapping of items to store in the bucket.
        """
        for key, value in items.items():
            self[key].put(value)

    def remove(
        self,
        keys: Iterable[KeyT],
        *,
        how: Callable[[LinkT], bool] | None = None,
    ) -> dict[KeyT, bool]:
        """Remove resources from the bucket.

        Args:
            keys: The keys to the resources.
            how: A function that removes the resource.

        Returns:
            A mapping of keys to whether they were removed.
        """
        return {key: self[key].remove(how=how) for key in keys}

    def __truediv__(self, key: KeyT) -> Self:
        try:
            return self.sub(key)
        except TypeError:
            return NotImplemented



from __future__ import annotations

import warnings
from collections.abc import Iterator
from contextlib import AbstractContextManager, ExitStack, contextmanager
from datetime import datetime
from typing import Any

import pandas as pd


def threadpoolctl_heuristic(item_contained_in_node: Any | None) -> bool:
    """Heuristic to determine if we should automatically set threadpoolctl.

    This is done by detecting if it's a scikit-learn `BaseEstimator` but this may
    be extended in the future.

    !!! tip

        The reason to have this heuristic is that when running scikit-learn, or any
        multithreaded model, in parallel, they will over subscribe to threads. This
        causes a significant performance hit as most of the time is spent switching
        thread contexts instead of work. This can be particularly bad for HPO where
        we are evaluating multiple models in parallel on the same system.

        The recommened thread count is 1 per core with no additional information to
        act upon.

    !!! todo

        This is potentially not an issue if running on multiple nodes of some cluster,
        as they do not share logical cores and hence do not clash.

    Args:
        item_contained_in_node: The item with which to base the heuristic on.

    Returns:
        Whether we should automatically set threadpoolctl.
    """
    if item_contained_in_node is None or not isinstance(item_contained_in_node, type):
        return False

    try:
        # NOTE: sklearn depends on threadpoolctl so it will be installed.
        from sklearn.base import BaseEstimator

        return issubclass(item_contained_in_node, BaseEstimator)
    except ImportError:
        return False


def parse_timestamp_object(timestamp: Any) -> datetime:
    """Parse a timestamp object, erring if it can't be parsed.

    Args:
        timestamp: The timestamp to parse.

    Returns:
        The parsed timestamp or `None` if it could not be parsed.
    """
    # Make sure we correctly set it's generated at if
    # we can
    match timestamp:
        case datetime():
            return timestamp
        case pd.Timestamp():
            return timestamp.to_pydatetime()
        case float() | int():
            return datetime.fromtimestamp(timestamp)
        case str():
            try:
                return datetime.fromisoformat(timestamp)
            except ValueError as e:
                raise ValueError(
                    f"Could not parse `str` type timestamp for '{timestamp}'."
                    " \nPlease provide a valid isoformat timestamp, e.g."
                    "'2021-01-01T00:00:00.000000'.",
                ) from e
        case _:
            raise TypeError(f"Could not parse {timestamp=} of type {type(timestamp)}.")


@contextmanager
def ignore_warnings() -> Iterator[None]:
    """Ignore warnings for the duration of the context manager."""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        yield


@contextmanager
def mutli_context(*managers: AbstractContextManager) -> Iterator:
    """Run multiple context managers at once."""
    with ExitStack() as stack:
        yield [stack.enter_context(m) for m in managers]



"""Implementation of the weighted ensemble procedure from Caruana et al. 2004.

???+ note "Reference"
    Ensemble selection from libraries of models

    Rich Caruana, Alexandru Niculescu-Mizil, Geoff Crew and Alex Ksikes

    ICML 2004

    https://dl.acm.org/doi/10.1145/1015330.1015432

    https://www.cs.cornell.edu/~caruana/ctp/ct.papers/caruana.icml04.icdm06long.pdf
"""
from __future__ import annotations

import logging
from collections import Counter
from collections.abc import Callable, Hashable, Iterable, Mapping
from typing import TYPE_CHECKING, TypeVar

import numpy as np

from amltk.randomness import as_rng

if TYPE_CHECKING:
    from amltk.types import Seed

logger = logging.getLogger(__name__)

# Values return by metric require that we can perform equality checks on them
T = TypeVar("T")
K = TypeVar("K", bound=Hashable)


def weighted_ensemble_caruana(
    *,
    model_predictions: Mapping[K, np.ndarray],
    targets: np.ndarray,
    size: int,
    metric: Callable[[np.ndarray, np.ndarray], T],
    select: Callable[[Iterable[T]], T],
    seed: Seed | None = None,
) -> tuple[dict[K, float], list[tuple[K, T]], np.ndarray]:
    """Calculate a weighted ensemble of `n` models.

    Args:
        model_predictions: Mapping from model id to predictions
        targets: The targets
        size: The size of the ensemble to create
        metric: The metric to use in calculating which models to add to the ensemble.
        select: Selects a models from the list based on the values of the metric on
            their predictions. Can return a single ID or a list of them, in which
            case a random selection will be made.
        seed: The seed to use for breaking ties

    Returns:
        A mapping from id's to it's weight in the ensemble and the trajectory.
    """
    if not size > 0:
        raise ValueError("`size` must be positive")

    if len(model_predictions) == 0:
        raise ValueError("`model_predictions` is empty")

    rng = as_rng(seed)
    predictions = list(model_predictions.values())

    dtype = predictions[0].dtype
    if np.issubdtype(dtype, np.integer):
        logger.warning(
            f"Predictions were {dtype=}, converting to np.float64 to"
            " allow for weighted ensemble procedure.",
        )
        dtype = np.float64

    # Current sum of predictions in the ensemble
    current = np.zeros_like(predictions[0], dtype=dtype)

    # Buffer where new models predictions are added to current to try them
    buffer = np.empty_like(predictions[0], dtype=dtype)

    ensemble: list[K] = []
    trajectory: list[tuple[K, T]] = []

    def value_if_added(_pred: np.ndarray) -> T:
        # Get the value if the model was added to the current set of predicitons
        np.add(current, _pred, out=buffer)
        np.multiply(buffer, (1.0 / float(len(ensemble) + 1)), out=buffer)

        return metric(targets, buffer)

    for _ in range(size):
        # Get the value if added for each model
        scores = {_id: value_if_added(pred) for _id, pred in model_predictions.items()}

        # Get the choices that produce the best value
        chosen_val = select(scores.values())

        choices = [_id for _id, score in scores.items() if score == chosen_val]
        choice = rng.choice(np.asarray(choices))

        # Add the predictions of the chosen model
        np.add(current, model_predictions[choice], out=current)

        # Record it's addition and the score of the ensemble with this
        # choice added
        ensemble.append(choice)
        trajectory.append((choice, chosen_val))

        # In the case of only one model, have calculated it's loss
        # and it's the only available model to add to the ensemble
        if len(model_predictions) == 1:
            ensemble *= size
            trajectory *= size
            break

    final = np.multiply(current, (1.0 / float(len(ensemble))))

    return (
        {_id: count / size for _id, count in Counter(ensemble).items()},
        trajectory,
        final,
    )



"""A portfolio in meta-learning is to a set (ordered or not) of configurations
that maximize some notion of coverage across datasets or tasks.
The intuition here is that this also means that any new dataset is also covered!

Suppose we have the given performances of some configurations across some datasets.
```python exec="true" source="material-block" result="python" title="Initial Portfolio"
import pandas as pd

performances = {
    "c1": [90, 60, 20, 10],
    "c2": [20, 10, 90, 20],
    "c3": [10, 20, 40, 90],
    "c4": [90, 10, 10, 10],
}
portfolio = pd.DataFrame(performances, index=["dataset_1", "dataset_2", "dataset_3", "dataset_4"])
print(portfolio)
```

If we could only choose `#!python k=3` of these configurations on some new given dataset, which ones would
you choose and in what priority?
Here is where we can apply [`portfolio_selection()`][amltk.metalearning.portfolio_selection]!

The idea is that we pick a subset of these algorithms that maximise some value of utility for
the portfolio. We do this by adding a single configuration from the entire set, 1-by-1 until
we reach `k`, beginning with the empty portfolio.

Let's see this in action!

```python exec="true" source="material-block" result="python" title="Portfolio Selection" hl_lines="12 13 14 15 16"
import pandas as pd
from amltk.metalearning import portfolio_selection

performances = {
    "c1": [90, 60, 20, 10],
    "c2": [20, 10, 90, 20],
    "c3": [10, 20, 40, 90],
    "c4": [90, 10, 10, 10],
}
portfolio = pd.DataFrame(performances, index=["dataset_1", "dataset_2", "dataset_3", "dataset_4"])

selected_portfolio, trajectory = portfolio_selection(
    portfolio,
    k=3,
    scaler="minmax"
)

print(selected_portfolio)
print()
print(trajectory)
```

The trajectory tells us which configuration was added at each time stamp along with the utility
of the portfolio with that configuration added. However we havn't specified how _exactly_ we defined the
utility of a given portfolio. We could define our own function to do so:

```python exec="true" source="material-block" result="python" title="Portfolio Selection Custom" hl_lines="12 13 14 20"
import pandas as pd
from amltk.metalearning import portfolio_selection

performances = {
    "c1": [90, 60, 20, 10],
    "c2": [20, 10, 90, 20],
    "c3": [10, 20, 40, 90],
    "c4": [90, 10, 10, 10],
}
portfolio = pd.DataFrame(performances, index=["dataset_1", "dataset_2", "dataset_3", "dataset_4"])

def my_function(p: pd.DataFrame) -> float:
    # Take the maximum score for each dataset and then take the mean across them.
    return p.max(axis=1).mean()

selected_portfolio, trajectory = portfolio_selection(
    portfolio,
    k=3,
    scaler="minmax",
    portfolio_value=my_function,
)

print(selected_portfolio)
print()
print(trajectory)
```

This notion of reducing across all configurations for a dataset and then aggregating these is common
enough that we can also directly just define these operations and we will perform the rest.

```python exec="true" source="material-block" result="python" title="Portfolio Selection With Reduction" hl_lines="17 18"
import pandas as pd
import numpy as np
from amltk.metalearning import portfolio_selection

performances = {
    "c1": [90, 60, 20, 10],
    "c2": [20, 10, 90, 20],
    "c3": [10, 20, 40, 90],
    "c4": [90, 10, 10, 10],
}
portfolio = pd.DataFrame(performances, index=["dataset_1", "dataset_2", "dataset_3", "dataset_4"])

selected_portfolio, trajectory = portfolio_selection(
    portfolio,
    k=3,
    scaler="minmax",
    row_reducer=np.max,  # This is actually the default
    aggregator=np.mean,  # This is actually the default
)

print(selected_portfolio)
print()
print(trajectory)
```
"""  # noqa: E501

from __future__ import annotations

from collections.abc import Callable, Hashable
from typing import TYPE_CHECKING, Literal, TypeVar

import numpy as np
import pandas as pd

from amltk.randomness import as_rng
from amltk.types import Seed, safe_isinstance

if TYPE_CHECKING:
    from sklearn.base import TransformerMixin

K = TypeVar("K", bound=Hashable)


def portfolio_selection(
    items: dict[K, pd.Series] | pd.DataFrame,
    k: int,
    *,
    row_reducer: Callable[[pd.Series], float] = np.max,
    aggregator: Callable[[pd.Series], float] = np.mean,
    portfolio_value: Callable[[pd.DataFrame], float] | None = None,
    maximize: bool = True,
    scaler: TransformerMixin | Literal["minmax"] | None = "minmax",
    with_replacement: bool = False,
    stop_if_worse: bool = False,
    seed: Seed | None = None,
) -> tuple[pd.DataFrame, pd.Series]:
    """Selects a portfolio of `k` items from `items`.

    A portfolio is a subset of the items, and is selected by maximizing the
    `portfolio_value` function in a greedy selection approach.

    At each iteration `0 <= i < k`, the `portfolio_value` function is calculated
    for the portfolio obtained by adding the `i`th item to the portfolio. The item
    that maximizes the `portfolio_value` function is then added to the portfolio for
    the next iteration.

    The `portfolio_function` can often be define by a row wise reduction
    (`row_reducer=`) followed by some aggregation over these reductions (`aggregator=`).
    You can also supply your own value function if desired (`portfolio_value=`).

    !!! example "A Single Iteration"

        This uses the `row_reducer=np.max` and `aggregator=np.mean` to calculate the
        value of a portfolio.

        In this case, we have 4 datasets and our current portfolio
        consists of `config_1` and `config_2`. We are going to calculate the value of
        adding `config_try` to the current best portfolio.

        ```python
                    | config_1 | config_2 | config_try
        dataset_1   |    1     |    0     |    0
        dataset_2   |    0     |   0.5    |    1
        dataset_3   |    0     |   0.5    |   0.5
        dataset_4   |    1     |    1     |    0
        ```

        Apply `row_reducer` to each row, in this case `np.max`

        ```python
                    |   max
        dataset_1   |    1
        dataset_2   |    1
        dataset_3   |   0.5
        dataset_4   |    1
        ```

        Apply `aggregator` to the reduced rows, in this case `np.mean`

        ```python
        portfolio_value = np.mean([1, 1, 0.5, 1]) # 0.875
        ```

    Args:
        items: A dictionary of items to select from.
        k: The number of items to select.
        row_reducer: A function to aggregate the rows of the portfolio.
            This is applied to a potential portfolio, for example to calculate
            the max score of all configs, for a given dataset (row).
        aggregator: A function to take all the single values reduced by `row_reducer`,
            and aggregate them into a final value for the portfolio.
        portfolio_value: A custom function to calculate the value of a portfolio.
            This will take precedence over `row_reducer` and `aggregator`.
        maximize: Whether to maximize or minimize the portfolio value.
        scaler: A scaler to use to scale the portfolio values. Is applied across
            the rows.
        with_replacement: Whether to select items with replacement.
        stop_if_worse: Whether to stop if the portfolio value is worse than the
            current best.
        seed: The seed to use for breaking ties.

    Returns:
        The final portfolio
        The trajectory, where the entry is the value once added to the portfolio.
    """
    if not (1 <= k < len(items)):
        raise ValueError(f"k must be in [1, {len(items)=})")

    all_portfolio = pd.DataFrame(items)

    # Normalize if needed
    if scaler is None:
        pass
    elif scaler == "minmax":
        min_maxs = all_portfolio.agg(["min", "max"], axis=1)

        mins = min_maxs["min"]
        maxs = min_maxs["max"]
        normalizer = maxs - mins

        # If everything is equal, we need to make sure the normalizing
        # doesn't do anything
        normalizer[normalizer == 0] = 1
        mins[normalizer == 0] = 0

        norm = lambda col: (col - mins) / normalizer
        all_portfolio: pd.DataFrame = all_portfolio.apply(norm, axis=0)  # type: ignore
    elif safe_isinstance(scaler, "TransformerMixin"):
        assert not isinstance(scaler, str)
        all_portfolio = scaler.fit_transform(all_portfolio.T).T
    else:
        raise ValueError(f"Invalid scaler: {scaler}")

    # Set up the portfolio value function
    if portfolio_value is None:
        portfolio_value = lambda _portfolio: float(
            aggregator(_portfolio.apply(row_reducer, axis=1)),
        )

    # Make a copy as we will del from it
    items = dict(items)
    rng = as_rng(seed)
    best = max if maximize else min

    # Running counters during the algorithm loop
    added_items: list[K] = []
    values: list[float] = []
    current_best: float = -np.inf if maximize else np.inf

    for _ in range(k):
        possible_portfolios = [(k, all_portfolio[[*added_items, k]]) for k in items]
        values_possible = {
            k: portfolio_value(possible_portfolio)
            for k, possible_portfolio in possible_portfolios
        }

        # This is the highest value we can get from a portfolio of the current size
        best_possible = best(values_possible.values())

        # If the best possible value of what we can do does not improve over the current
        #    best portfolio, stop (if enabled)
        if stop_if_worse and current_best == best(best_possible, current_best):
            break

        current_best = best_possible

        # Possible get multiple best choices, we choose one at random if so
        best_keys = [k for k, v in values_possible.items() if v == best_possible]
        best_key = (
            best_keys[0] if len(best_keys) == 1 else rng.choice(best_keys)  # type: ignore
        )

        # We found something better, add it in
        added_items.append(best_key)
        values.append(best_possible)

        if not with_replacement:
            del items[best_key]

    # Rename the columns of the portfolio to be the keys
    return all_portfolio[added_items], pd.Series(values, index=added_items)



"""Module to measure memory."""
from __future__ import annotations

from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, Literal
from typing_extensions import override

import numpy as np
import pandas as pd
import psutil

from amltk._functional import dict_get_not_none

if TYPE_CHECKING:
    from pandas._libs.missing import NAType


@dataclass
class Memory:
    """A timer for measuring the time between two events.

    Attributes:
        start_vms: The virtual memory size at the start of the interval.
        start_rss: The resident set size at the start of the interval.
        unit: The unit of the memory.
    """

    start_vms: float
    start_rss: float
    unit: Memory.Unit | NAType

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> Memory.Interval:
        """Create a memory interval from a dictionary.

        Args:
            d: The dictionary to create from.

        Returns:
            The memory interval.
        """
        return Memory.Interval(
            start_vms=dict_get_not_none(d, "start_vms", np.nan),
            start_rss=dict_get_not_none(d, "start_rss", np.nan),
            end_vms=dict_get_not_none(d, "end_vms", np.nan),
            end_rss=dict_get_not_none(d, "end_rss", np.nan),
            unit=Memory.Unit.from_str(dict_get_not_none(d, "unit", pd.NA)),
        )

    @classmethod
    def na(cls) -> Memory.Interval:
        """Create a memory interval that represents NA."""
        return Memory.Interval(
            start_vms=np.nan,
            end_vms=np.nan,
            start_rss=np.nan,
            end_rss=np.nan,
            unit=pd.NA,
        )

    @classmethod
    def convert(
        cls,
        x: float,
        frm: Memory.Unit | NAType | Literal["B", "KB", "MB", "GB"] = "B",
        to: Memory.Unit | NAType | Literal["B", "KB", "MB", "GB"] = "B",
    ) -> float:
        """Convert a value from one unit to another.

        Args:
            x: The value to convert.
            frm: The unit of the value.
            to: The unit to convert to.

        Returns:
            The converted value.
        """
        if frm == to:
            return x

        _frm = cls.Unit.from_str(frm) if isinstance(frm, str) else frm
        _to = cls.Unit.from_str(to) if isinstance(to, str) else to
        return x * _CONVERSION[_frm] / _CONVERSION[_to]

    @classmethod
    @contextmanager
    def measure(
        cls,
        unit: Memory.Unit | Literal["B", "KB", "MB", "GB"] = "B",
    ) -> Iterator[Memory.Interval]:
        """Measure the memory used by a block of code.

        Args:
            unit: The unit of memory to use.

        Yields:
            The Memory Interval. The start and end memory will not be
            valid until the context manager is exited.

        """
        mem = cls.start(unit=unit)

        interval = Memory.na()
        interval.unit = mem.unit

        try:
            yield interval
        finally:
            _interval = mem.stop()
            interval.start_vms = _interval.start_vms
            interval.end_vms = _interval.end_vms
            interval.start_rss = _interval.start_rss
            interval.end_rss = _interval.end_rss

    @classmethod
    def start(
        cls,
        unit: Memory.Unit | Literal["B", "KB", "MB", "GB"] = "B",
    ) -> Memory:
        """Start a memory tracker.

        Args:
            unit: The unit of memory to use (bytes).

        Returns:
            The Memory tracker.
        """
        proc = psutil.Process()
        info = proc.memory_info()

        return Memory(
            start_vms=Memory.convert(info.vms, frm=Memory.Unit.BYTES, to=unit),
            start_rss=Memory.convert(info.rss, frm=Memory.Unit.BYTES, to=unit),
            unit=unit if isinstance(unit, Memory.Unit) else Memory.Unit.from_str(unit),
        )

    def stop(self) -> Memory.Interval:
        """Stop the memory tracker.

        Returns:
            The memory interval.
        """
        proc = psutil.Process()
        info = proc.memory_info()

        return Memory.Interval(
            start_vms=self.start_vms,
            start_rss=self.start_rss,
            end_vms=Memory.convert(info.vms, frm=Memory.Unit.BYTES, to=self.unit),
            end_rss=Memory.convert(info.rss, frm=Memory.Unit.BYTES, to=self.unit),
            unit=self.unit,
        )

    @classmethod
    def usage(
        cls,
        kind: Literal["rss", "vms"] = "vms",
        unit: Memory.Unit | Literal["B", "KB", "MB", "GB"] = "B",
    ) -> float:
        """Get the memory used.

        Args:
            kind: The type of memory to measure.
            unit: The unit of memory to use.

        Returns:
            The memory used.
        """
        proc = psutil.Process()
        _kind = kind.lower()
        if _kind == "rss":
            return Memory.convert(
                proc.memory_info().rss,
                frm=Memory.Unit.BYTES,
                to=unit,
            )

        if _kind == "vms":
            return Memory.convert(
                proc.memory_info().vms,
                frm=Memory.Unit.BYTES,
                to=unit,
            )

        raise ValueError(f"Unknown memory type: {kind}")

    @dataclass
    class Interval:
        """A class for representing a time interval.

        Attributes:
            start_vms: The virtual memory size at the start of the interval.
            start_rss: The resident set size at the start of the interval.
            end_vms: The virtual memory size at the end of the interval.
            end_rss: The resident set size at the end of the interval.
            unit: The unit of memory.
        """

        start_vms: float
        start_rss: float
        end_vms: float
        end_rss: float
        unit: Memory.Unit | NAType

        @property
        def vms_used(self) -> float:
            """The amount of vms memory used in the interval.

            !!!warning

                This does not track peak memory usage. This will
                only give the difference between the start and end
                of the interval.
            """
            return self.end_vms - self.start_vms

        @property
        def rss_used(self) -> float:
            """The amount of rss memory used in the interval.

            !!!warning

                This does not track peak memory usage. This will
                only give the difference between the start and end
                of the interval.
            """
            return self.end_rss - self.start_rss

        def to_unit(self, unit: Memory.Unit) -> Memory.Interval:
            """Return the memory used in a different unit.

            Args:
                unit: The unit to convert to.

            Returns:
                The memory used in the new unit.
            """
            if self.unit == unit:
                return self

            return Memory.Interval(
                start_vms=Memory.convert(self.start_vms, self.unit, unit),
                end_vms=Memory.convert(self.end_vms, self.unit, unit),
                start_rss=Memory.convert(self.start_rss, self.unit, unit),
                end_rss=Memory.convert(self.end_rss, self.unit, unit),
                unit=unit,
            )

        def to_dict(self, *, prefix: str = "") -> dict[str, Any]:
            """Convert the time interval to a dictionary."""
            return {
                f"{prefix}start_vms": self.start_vms,
                f"{prefix}end_vms": self.end_vms,
                f"{prefix}diff_vms": self.vms_used,
                f"{prefix}start_rss": self.start_rss,
                f"{prefix}end_rss": self.end_rss,
                f"{prefix}diff_rss": self.rss_used,
                f"{prefix}unit": self.unit,
            }

    class Unit(str, Enum):
        """An enum for the units of time."""

        BYTES = "bytes"
        KILOBYTES = "kilobytes"
        MEGABYTES = "megabytes"
        GIGABYTES = "gigabytes"

        @override
        def __str__(self) -> str:
            return self.name.lower()

        @override
        def __repr__(self) -> str:
            return self.name.lower()

        @classmethod
        def from_str(cls, s: Any) -> Memory.Unit | NAType:
            """Convert a string to a unit."""
            if isinstance(s, Memory.Unit):
                return s

            if isinstance(s, str):
                _mapping = {
                    "bytes": Memory.Unit.BYTES,
                    "b": Memory.Unit.BYTES,
                    "kilobytes": Memory.Unit.KILOBYTES,
                    "kb": Memory.Unit.KILOBYTES,
                    "megabytes": Memory.Unit.MEGABYTES,
                    "mb": Memory.Unit.MEGABYTES,
                    "gigabytes": Memory.Unit.GIGABYTES,
                    "gb": Memory.Unit.GIGABYTES,
                }
                return _mapping.get(s.lower(), pd.NA)

            return pd.NA


_CONVERSION: Mapping[Memory.Unit | NAType, float] = {
    Memory.Unit.BYTES: 1,
    Memory.Unit.KILOBYTES: 1024,
    Memory.Unit.MEGABYTES: 1024**2,
    Memory.Unit.GIGABYTES: 1024**3,
    pd.NA: np.nan,
}



"""Dask Jobqueue Executors.

These are essentially wrappers around the dask_jobqueue classes.
We use them to provide a consistent interface for all the different
jobqueue implementations and get access to their executors.

!!! example "Documentation from `dask_jobqueue`"

    See the [dask jobqueue documentation](https://jobqueue.dask.org/en/latest/) specifically:

    * [Example deployments](https://jobqueue.dask.org/en/latest/examples.html#example-deployments)
    * [Tips and Tricks](https://jobqueue.dask.org/en/latest/advanced-tips-and-tricks.html)
    * [Debugging](https://jobqueue.dask.org/en/latest/debug.html)

"""
# ruff: noqa: N802, E501
from __future__ import annotations

import logging
import pprint
from collections.abc import Callable, Iterable, Iterator
from concurrent.futures import Executor, Future
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeAlias, TypeVar
from typing_extensions import ParamSpec, Self, override

from dask_jobqueue import (
    HTCondorCluster,
    JobQueueCluster,
    LSFCluster,
    MoabCluster,
    OARCluster,
    PBSCluster,
    SGECluster,
    SLURMCluster,
)

if TYPE_CHECKING:
    from distributed.cfexecutor import ClientExecutor

R = TypeVar("R")
P = ParamSpec("P")
_JQC = TypeVar("_JQC", bound=JobQueueCluster)

DJQ_NAMES: TypeAlias = Literal["slurm", "htcondor", "lsf", "oar", "pbs", "sge", "moab"]

logger = logging.getLogger(__name__)


class DaskJobqueueExecutor(Executor, Generic[_JQC]):
    """A concurrent.futures Executor that executes tasks on a dask_jobqueue cluster."""

    def __init__(
        self,
        cluster: _JQC,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
    ):
        """Initialize a DaskJobqueueExecutor.


        !!! note "Implementations"

            Prefer to use the class methods to create an instance of this class.

            * [`DaskJobqueueExecutor.SLURM()`][amltk.scheduling.executors.dask_jobqueue.DaskJobqueueExecutor.SLURM]
            * [`DaskJobqueueExecutor.HTCondor()`][amltk.scheduling.executors.dask_jobqueue.DaskJobqueueExecutor.HTCondor]
            * [`DaskJobqueueExecutor.LSF()`][amltk.scheduling.executors.dask_jobqueue.DaskJobqueueExecutor.LSF]
            * [`DaskJobqueueExecutor.OAR()`][amltk.scheduling.executors.dask_jobqueue.DaskJobqueueExecutor.OAR]
            * [`DaskJobqueueExecutor.PBS()`][amltk.scheduling.executors.dask_jobqueue.DaskJobqueueExecutor.PBS]
            * [`DaskJobqueueExecutor.SGE()`][amltk.scheduling.executors.dask_jobqueue.DaskJobqueueExecutor.SGE]
            * [`DaskJobqueueExecutor.Moab()`][amltk.scheduling.executors.dask_jobqueue.DaskJobqueueExecutor.Moab]

        Args:
            cluster: The implementation of a
                [dask_jobqueue.JobQueueCluster](https://jobqueue.dask.org/en/latest/api.html).
            n_workers: The number of workers to maximally adapt to on the cluster.
            adaptive: Whether to use the adaptive scaling of the cluster or fixed allocate all workers.
                This will specifically use the
                [dask_jobqueue.SLURMCluster.adapt](https://jobqueue.dask.org/en/latest/index.html?highlight=adapt#adaptivity)
                method to dynamically scale the cluster to the number of workers specified.
            submit_command: To overwrite the submission command if necessary.
            cancel_command: To overwrite the cancel command if necessary.
        """
        super().__init__()
        self.cluster = cluster
        self.adaptive = adaptive
        if submit_command:
            self.cluster.job_cls.submit_command = submit_command  # type: ignore

        if cancel_command:
            self.cluster.job_cls.cancel_command = cancel_command  # type: ignore

        if adaptive:
            self.cluster.adapt(minimum=0, maximum=n_workers)
        else:
            self.cluster.scale(n_workers)

        self.n_workers = n_workers
        self._client = self.cluster.get_client()
        self.executor: ClientExecutor = self._client.get_executor()

    @override
    def __enter__(self) -> Self:
        configuration = {
            "header": self.cluster.job_header,
            "script": self.cluster.job_script(),
            "job_name": self.cluster.job_name,
        }
        config_str = pprint.pformat(configuration)
        logger.debug(f"Launching script with configuration:\n {config_str}")
        self.executor.__enter__()
        return self

    @override
    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        self.executor.__exit__(*args, **kwargs)
        self._client.close()

    @override
    def submit(
        self,
        fn: Callable[P, R],
        /,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Future[R]:
        """See [concurrent.futures.Executor.submit][]."""
        future = self.executor.submit(fn, *args, **kwargs)
        assert isinstance(future, Future)
        return future

    @override
    def map(
        self,
        fn: Callable[..., R],
        *iterables: Iterable,
        timeout: float | None = None,
        chunksize: int = 1,
    ) -> Iterator[R]:
        """See [concurrent.futures.Executor.map][]."""
        return self.executor.map(  # type: ignore
            fn,
            *iterables,
            timeout=timeout,
            chunksize=chunksize,
        )

    @override
    def shutdown(
        self,
        wait: bool = True,
        **kwargs: Any,
    ) -> None:
        """See [concurrent.futures.Executor.shutdown][]."""
        self.executor.shutdown(wait=wait, **kwargs)

    @classmethod
    def SLURM(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> DaskJobqueueExecutor[SLURMCluster]:
        """Create a DaskJobqueueExecutor for a SLURM cluster.

        See the [dask_jobqueue.SLURMCluster documentation][dask_jobqueue.SLURMCluster] for
        more information on the available keyword arguments.
        """
        return cls(  # type: ignore
            SLURMCluster(**kwargs),
            submit_command=submit_command,
            cancel_command=cancel_command,
            n_workers=n_workers,
            adaptive=adaptive,
        )

    @classmethod
    def HTCondor(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> DaskJobqueueExecutor[HTCondorCluster]:
        """Create a DaskJobqueueExecutor for a HTCondor cluster.

        See the [dask_jobqueue.HTCondorCluster documentation][dask_jobqueue.HTCondorCluster] for
        more information on the available keyword arguments.
        """
        return cls(  # type: ignore
            HTCondorCluster(**kwargs),
            submit_command=submit_command,
            cancel_command=cancel_command,
            adaptive=adaptive,
            n_workers=n_workers,
        )

    @classmethod
    def LSF(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> DaskJobqueueExecutor[LSFCluster]:
        """Create a DaskJobqueueExecutor for a LSF cluster.

        See the [dask_jobqueue.LSFCluster documentation][dask_jobqueue.LSFCluster] for
        more information on the available keyword arguments.
        """
        return cls(  # type: ignore
            LSFCluster(**kwargs),
            submit_command=submit_command,
            cancel_command=cancel_command,
            adaptive=adaptive,
            n_workers=n_workers,
        )

    @classmethod
    def OAR(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> DaskJobqueueExecutor[OARCluster]:
        """Create a DaskJobqueueExecutor for a OAR cluster.

        See the [dask_jobqueue.OARCluster documentation][dask_jobqueue.OARCluster] for
        more information on the available keyword arguments.
        """
        return cls(  # type: ignore
            OARCluster(**kwargs),
            submit_command=submit_command,
            cancel_command=cancel_command,
            adaptive=adaptive,
            n_workers=n_workers,
        )

    @classmethod
    def PBS(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> DaskJobqueueExecutor[PBSCluster]:
        """Create a DaskJobqueueExecutor for a PBS cluster.

        See the [dask_jobqueue.PBSCluster documentation][dask_jobqueue.PBSCluster] for
        more information on the available keyword arguments.
        """
        return cls(  # type: ignore
            PBSCluster(**kwargs),
            submit_command=submit_command,
            cancel_command=cancel_command,
            adaptive=adaptive,
            n_workers=n_workers,
        )

    @classmethod
    def SGE(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> DaskJobqueueExecutor[SGECluster]:
        """Create a DaskJobqueueExecutor for a SGE cluster.

        See the [dask_jobqueue.SGECluster documentation][dask_jobqueue.SGECluster] for
        more information on the available keyword arguments.
        """
        return cls(  # type: ignore
            SGECluster(**kwargs),
            submit_command=submit_command,
            cancel_command=cancel_command,
            adaptive=adaptive,
            n_workers=n_workers,
        )

    @classmethod
    def Moab(
        cls,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> DaskJobqueueExecutor[MoabCluster]:
        """Create a DaskJobqueueExecutor for a Moab cluster.

        See the [dask_jobqueue.MoabCluster documentation][dask_jobqueue.MoabCluster] for
        more information on the available keyword arguments.
        """
        return cls(  # type: ignore
            MoabCluster(**kwargs),
            submit_command=submit_command,
            cancel_command=cancel_command,
            adaptive=adaptive,
            n_workers=n_workers,
        )

    @classmethod
    def from_str(
        cls,
        name: DJQ_NAMES,
        *,
        n_workers: int,
        adaptive: bool = False,
        submit_command: str | None = None,
        cancel_command: str | None = None,
        **kwargs: Any,
    ) -> DaskJobqueueExecutor:
        """Create a DaskJobqueueExecutor using a string lookup.

        Args:
            name: The name of cluster to create, must be one of
                ["slurm", "htcondor", "lsf", "oar", "pbs", "sge", "moab"].
            n_workers: The number of workers to maximally adapt to on the cluster.
            adaptive: Whether to use the adaptive scaling of the cluster or fixed allocate all workers.
                This will specifically use the
                [dask_jobqueue.SLURMCluster.adapt](https://jobqueue.dask.org/en/latest/index.html?highlight=adapt#adaptivity)
                method to dynamically scale the cluster to the number of workers specified.
            submit_command: Overwrite the submit command of workers if necessary.
            cancel_command: Overwrite the cancel command of workers if necessary.
            kwargs: The keyword arguments to pass to the cluster constructor.

        Raises:
            KeyError: If `name` is not one of the supported cluster types.

        Returns:
            A DaskJobqueueExecutor for the requested cluster type.
        """
        methods = {
            "slurm": cls.SLURM,
            "htcondor": cls.HTCondor,
            "lsf": cls.LSF,
            "oar": cls.OAR,
            "pbs": cls.PBS,
            "sge": cls.SGE,
            "moab": cls.Moab,
        }
        method = methods.get(name.lower())
        if method is None:
            raise KeyError(
                f"Unknown cluster name: {name}, must be from {list(methods)}",
            )

        return method(
            n_workers=n_workers,
            submit_command=submit_command,
            cancel_command=cancel_command,
            adaptive=adaptive,
            **kwargs,
        )
